#include <sys/time.h>
#include <unistd.h>
#include <linux/input.h>

#include "quakedef.h"
#include "errno.h"

extern int cores;
extern int ir;
extern int kb;

/*
===============================================================================

FILE IO

===============================================================================
*/

qboolean			isDedicated;

#define MAX_HANDLES             10
FILE    *sys_handles[MAX_HANDLES];

int             findhandle (void)
{
	int             i;
	
	for (i=1 ; i<MAX_HANDLES ; i++)
		if (!sys_handles[i])
			return i;
	Sys_Error ("out of handles");
	return -1;
}

/*
================
filelength
================
*/
int filelength (FILE *f)
{
	int             pos;
	int             end;

	pos = ftell (f);
	fseek (f, 0, SEEK_END);
	end = ftell (f);
	fseek (f, pos, SEEK_SET);

	return end;
}

int Sys_FileOpenRead (char *path, int *hndl)
{
	FILE    *f;
	int             i;
	
	i = findhandle ();

	f = fopen(path, "rb");
	if (!f)
	{
		*hndl = -1;
		return -1;
	}
	sys_handles[i] = f;
	*hndl = i;
	
	return filelength(f);
}

int Sys_FileOpenWrite (char *path)
{
	FILE    *f;
	int             i;
	
	i = findhandle ();

	f = fopen(path, "wb");
	if (!f)
		Sys_Error ("Error opening %s: %s", path,strerror(errno));
	sys_handles[i] = f;
	
	return i;
}

void Sys_FileClose (int handle)
{
	fclose (sys_handles[handle]);
	sys_handles[handle] = NULL;
}

void Sys_FileSeek (int handle, int position)
{
	fseek (sys_handles[handle], position, SEEK_SET);
}

int Sys_FileRead (int handle, void *dest, int count)
{
	return fread (dest, 1, count, sys_handles[handle]);
}

int Sys_FileWrite (int handle, void *data, int count)
{
	return fwrite (data, 1, count, sys_handles[handle]);
}

int     Sys_FileTime (char *path)
{
	FILE    *f;
	
	f = fopen(path, "rb");
	if (f)
	{
		fclose(f);
		return 1;
	}
	
	return -1;
}

void Sys_mkdir (char *path)
{
}


/*
===============================================================================

SYSTEM IO

===============================================================================
*/

void Sys_MakeCodeWriteable (unsigned long startaddr, unsigned long length)
{
}


void Sys_Error (char *error, ...)
{
	va_list         argptr;

	printf ("Sys_Error: ");   
	va_start (argptr,error);
	vprintf (error,argptr);
	va_end (argptr);
	printf ("\n");

	exit (1);
}

void Sys_Printf (char *fmt, ...)
{
	va_list         argptr;
	
	va_start (argptr,fmt);
	vprintf (fmt,argptr);
	va_end (argptr);
}

void Sys_Quit (void)
{
	exit (0);
}

double Sys_FloatTime (void)
{
    struct timeval  tp;
    struct timezone tzp;
    static int      secbase = 0;

    gettimeofday( &tp, &tzp );

    if( !secbase ) {
        secbase = tp.tv_sec;
        return tp.tv_usec/1000000.0;
    }

    return (tp.tv_sec - secbase) + tp.tv_usec/1000000.0;
}


char *Sys_ConsoleInput (void)
{
	return NULL;
}

void Sys_Sleep (void)
{
}


static int xlatekey( int sym )
{
    int rc;

    // ir only remap
    switch( sym ) {
    case  1: rc = '1';          break;
    case  2: rc = '2';          break;
    case  3: rc = '3';          break;
    case  4: rc = '4';          break;
    case  5: rc = '5';          break;
    case  6: rc = '6';          break;
    case  7: rc = '7';          break;
    case  8: rc = '8';          break;
    case  9: rc = '9';          break;
    case 10: rc = '0';          break;
    case 11: rc = 'y';          break; // *
    case 12: rc = 'n';          break; // #
    case 31: rc = K_LEFTARROW;  break;
    case 32: rc = K_RIGHTARROW; break;
    case 30: rc = K_DOWNARROW;  break;
    case 29: rc = K_UPARROW;    break;
    case 33: rc = K_ENTER;      break; // OK
    case 36: rc = K_TAB;        break; // Selfview
    case 37: rc = K_CTRL;       break; // Layout
    case 38: rc = K_ESCAPE;     break; // Cancel
    case 40: rc = ' ';          break; // Directory
    case 49: rc = K_F11;        break; // Presets
    default: rc = 0;            break;
    }

    return rc;
}

// need to xlate KEY_ to K_ here if we get kb support
static int do_kb( void )
{
    int rc;
    struct input_event ev;

    rc = read( kb, &ev, sizeof(ev) );
    if( rc != sizeof(ev) ) return 0;

    Key_Event( ev.code, ev.value != 0 );

    return 1;
}

int currkey = 0;
double downtime = 0;

static int do_ir( void )
{
    int rc;
    struct input_event ev;
    int doomkey;

    // any keys waiting?
    rc = read( ir, &ev, sizeof(ev) );
    if( rc != sizeof(ev) ) {
        if( downtime != 0 && Sys_FloatTime() >= downtime + 0.1f ) {
            Key_Event( currkey, false );
            currkey = 0;
            downtime = 0;
        }
        return 0;
    }

    doomkey = xlatekey( ev.code );
//    printf( "%f key %d %d\n", Sys_FloatTime(), ev.code, doomkey );
    if( doomkey == 0 ) return 1;


    // same key as before?
    if( doomkey == currkey ) {
        downtime = Sys_FloatTime();
        return 1;
    }

    // key switched without timeout?
    if( currkey != doomkey ) {
        Key_Event( currkey, false );
    }

    // send new keypress
    Key_Event( doomkey, true );
    currkey = doomkey;
    downtime = Sys_FloatTime();

    return 1;

}

void Sys_SendKeyEvents (void)
{
    if( kb >= 0 ) {
        while( do_kb() )
            ;
        return;
    }
    if( ir >= 0 ) {
        while( do_ir() )
            ;
    }
}

void Sys_HighFPPrecision (void)
{
}

void Sys_LowFPPrecision (void)
{
}

//=============================================================================

int quake_main (int argc, char *argv[])
{
    double time, oldtime, newtime;
    static quakeparms_t parms;
    extern int vcrFile;
    extern int recording;

	parms.memsize = (256*1024*1024)/cores;
	parms.membase = malloc (parms.memsize);
	parms.basedir = ".";

	COM_InitArgv (argc, argv);

	parms.argc = com_argc;
	parms.argv = com_argv;

	printf ("Host_Init\n");
	Host_Init (&parms);

    oldtime = Sys_FloatTime () - 0.1;

    while( 1 ) {
        // find time spent rendering last frame
        newtime = Sys_FloatTime();
        time = newtime - oldtime;

        if( cls.state == ca_dedicated ) {
            // play vcrfiles at max speed
            if( time < sys_ticrate.value && (vcrFile == -1 || recording) ) {
                usleep(1);
                continue;       // not time to run a server only tic yet
            }
            time = sys_ticrate.value;
        }

        if( time > sys_ticrate.value*2 )
            oldtime = newtime;
        else
            oldtime += time;

        Host_Frame( time );

    }

}


