// MultiQuake video stuff for TILE-Gx36.
// Nils Liaaen Corneliusen 2014
// https://www.ignorantus.com/

#include <stdio.h>
#include <stdint.h>
#include <unistd.h>

#include <postinc.h>
#include <tmc/sync.h>

#include "quakedef.h"
#include "d_local.h"

viddef_t vid;   // global video state
unsigned short d_8to16table[256];
unsigned       d_8to24table[256];


extern int cores;
extern int measure;
extern int multiply;
extern uint32_t *myscreen;
extern int mystride;

uint8_t  *vid_buffer = NULL;
uint8_t  *surfcache  = NULL;
short    *zbuffer    = NULL;

static uint32_t pal32[256];

void VID_SetPalette( unsigned char *palette )
{
    int i;

    for( i = 0; i < 256; ++i ) {
        uint64_t r,g,b;
        r = gammatable[*palette++];
        g = gammatable[*palette++];
        b = gammatable[*palette++];
        pal32[i] = __insn_v2int_l( __insn_v1int_l( 0, r ), __insn_v1int_l( g, b ) );
    }

}

void VID_ShiftPalette( unsigned char *palette )
{
    VID_SetPalette( palette );
}

void VID_Init( unsigned char *palette )
{
    int surfsize;
    surfsize   = (256*1024*1024)/cores;
    surfcache  = malloc( surfsize );
    vid_buffer = malloc( SCREENWIDTH*SCREENHEIGHT );
    zbuffer    = malloc( SCREENWIDTH*SCREENHEIGHT*sizeof(short) );

    vid.maxwarpwidth = vid.width = vid.conwidth = SCREENWIDTH;
    vid.maxwarpheight = vid.height = vid.conheight = SCREENHEIGHT;
    vid.aspect = 1.0;
    vid.numpages = 1;
    vid.colormap = host_colormap;
    vid.fullbright = 256 - LittleLong (*((int *)vid.colormap + 2048));
    vid.buffer = vid.conbuffer = vid_buffer;
    vid.rowbytes = vid.conrowbytes = SCREENWIDTH;
	
    d_pzbuffer = zbuffer;
    D_InitCaches( surfcache, surfsize );

    VID_SetPalette(palette);
}

void VID_Shutdown( void )
{

}


// Convert from palette based picture and scale to destination

static void scale1( uint8_t *src, uint32_t *pal, int srcstride, uint32_t *dst, int dststride )
{
    int x, y;
    uint32_t dstadd;

    dstadd = dststride-SCREENWIDTH;

    for( y = 0; y < SCREENHEIGHT; y++ ) {

        for( x = 0; x < SCREENWIDTH; x += 8 ) {
            uint64_t srcw;
            srcw = __insn_ld_add( src, 8 );
            __insn_st_add( dst, __insn_v4int_l( pal[__insn_bfextu( srcw,  8, 15 )], pal[__insn_bfextu( srcw,  0,  7 )] ), 8 );
            __insn_st_add( dst, __insn_v4int_l( pal[__insn_bfextu( srcw, 24, 31 )], pal[__insn_bfextu( srcw, 16, 23 )] ), 8 );
            __insn_st_add( dst, __insn_v4int_l( pal[__insn_bfextu( srcw, 40, 47 )], pal[__insn_bfextu( srcw, 32, 39 )] ), 8 );
            __insn_st_add( dst, __insn_v4int_l( pal[__insn_bfextu( srcw, 56, 63 )], pal[__insn_bfextu( srcw, 48, 55 )] ), 8 );
        }

        dst += dstadd;

    }

}

// Based on pseudocode from
// http://en.wikipedia.org/wiki/Image_scaling#EPX.2FScale2.C3.97.2FAdvMAME2.C3.97
//   A    --\ 1 2
// C P B  --/ 3 4
//   D
static void scale2_epx( uint8_t *src, uint32_t *pal, int srcstride, uint32_t *dst, int dststride )
{
    uint8_t  *src0, *src1, *src2;
    uint32_t *dst0, *dst1;
    int x, y;
    uint32_t a,b,c,d,p;
    uint32_t d1,d2,d3,d4;
    uint32_t dstadd;

    dst0 = dst;
    dst1 = dst + dststride;

    src0 = src;
    src1 = src;
    src2 = src + srcstride;

    dstadd = dststride*2-SCREENWIDTH*2;

    for( y = 0; y < SCREENHEIGHT; y++ ) {

        for( x = 0; x < SCREENWIDTH; x++ ) {

            a =                                     pal[src0[x]  ];
            c = x == 0             ? pal[src1[x]] : pal[src1[x-1]];
            p =                                     pal[src1[x]  ];
            b = x == SCREENWIDTH-1 ? pal[src1[x]] : pal[src1[x+1]];
            d =                                     pal[src2[x]  ];

            d1 = c == a && c != d && a != b ? a : p;
            d2 = a == b && a != c && b != d ? b : p;
            d4 = b == d && b != a && d != c ? d : p;
            d3 = d == c && d != b && c != a ? c : p;

            *dst0++ = d1; *dst0++ = d2;
            *dst1++ = d3; *dst1++ = d4;

        }

        dst0 += dstadd;
        dst1 += dstadd;

        src0 = src1;
        src1 = src2;
        src2 = y == SCREENHEIGHT-2 ? src1 : src2 + srcstride;
    }
}

// Based on pseudocode from
// http://en.wikipedia.org/wiki/Image_scaling#Scale3.C3.97.2FAdvMAME3.C3.97
// A B C --\  1 2 3
// D E F    > 4 5 6
// G H I --/  7 8 9
static void scale3_epx( uint8_t *src, uint32_t *pal, int srcstride, uint32_t *dst, int dststride )
{
    uint8_t  *src0, *src1, *src2;
    uint32_t *dst0, *dst1, *dst2;
    int x,y;
    uint32_t a,b,c,d,e,f,g,h,i;
    uint32_t d1,d2,d3,d4,d5,d6,d7,d8,d9;
    uint32_t dstadd;

    src0 = src;
    src1 = src;
    src2 = src + srcstride;

    dst0 = dst;
    dst1 = dst +   dststride;
    dst2 = dst + 2*dststride;

    dstadd = dststride*3-SCREENWIDTH*3;

    for( y = 0; y < SCREENHEIGHT; y++ ) {

        for( x = 0; x < SCREENWIDTH; x++ ) {
            int xl, xr;

            xl = x == 0             ? x : x-1;
            xr = x == SCREENWIDTH-1 ? x : x+1;

            a = pal[src0[xl]]; b = pal[src0[x]]; c = pal[src0[xr]];
            d = pal[src1[xl]]; e = pal[src1[x]]; f = pal[src1[xr]];
            g = pal[src2[xl]]; h = pal[src2[x]]; i = pal[src2[xr]];

            d1 =  d == b && d != h && b != f                                                        ? d : e;
            d2 = (d == b && d != h && b != f && e != c ) || (b == f && b != d && f != h && e != a)  ? b : e;
            d3 =                                             b == f && b != d && f != h             ? f : e;
            d4 = (h == d && h != f && d != b && e != a ) || (d == b && d != h && b != f && e != g ) ? d : e;
            d5 =                                                                                          e;
            d6 = (b == f && b != d && f != h && e != i ) || (f == h && f != b && h != d && e != c)  ? f : e;
            d7 =                                             h == d && h != f && d != b             ? d : e;
            d8 = (f == h && f != b && h != d && e != g ) || (h == d && h != f && d != b && e != i)  ? h : e;
            d9 =  f == h && f != b && h != d                                                        ? f : e;

            *dst0++ = d1; *dst0++ = d2; *dst0++ = d3;
            *dst1++ = d4; *dst1++ = d5; *dst1++ = d6;
            *dst2++ = d7; *dst2++ = d8; *dst2++ = d9;

        }

        dst0 += dstadd;
        dst1 += dstadd;
        dst2 += dstadd;

        src0 = src1;
        src1 = src2;
        src2 = y == SCREENHEIGHT-2 ? src1 : src2 + srcstride;
    }
}

static void scale4( uint8_t *src, uint32_t *pal, int srcstride, uint32_t *dst, int dststride )
{
    int x, y;
    uint32_t dstadd;
    uint32_t *dst0,*dst1,*dst2,*dst3;

    dst0 = dst;
    dst1 = dst +   dststride;
    dst2 = dst + 2*dststride;
    dst3 = dst + 3*dststride;

    dstadd = dststride*4-SCREENWIDTH*4;

    for( y = 0; y < SCREENHEIGHT; y++ ) {

        for( x = 0; x < SCREENWIDTH; x += 8 ) {
            uint64_t d,d00,d11,d22,d33,d44,d55,d66,d77;
            uint64_t srcw;

            srcw = __insn_ld_add( src, 8 );

            d = pal[__insn_bfextu( srcw,  0,  7 )]; d00 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw,  8, 15 )]; d11 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw, 16, 23 )]; d22 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw, 24, 31 )]; d33 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw, 32, 39 )]; d44 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw, 40, 47 )]; d55 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw, 48, 55 )]; d66 = __insn_v4int_l( d, d );
            d = pal[__insn_bfextu( srcw, 56, 63 )]; d77 = __insn_v4int_l( d, d );

            __insn_st_add( dst0, d00, 8 ); __insn_st_add( dst0, d00, 8 );
            __insn_st_add( dst0, d11, 8 ); __insn_st_add( dst0, d11, 8 );
            __insn_st_add( dst0, d22, 8 ); __insn_st_add( dst0, d22, 8 );
            __insn_st_add( dst0, d33, 8 ); __insn_st_add( dst0, d33, 8 );
            __insn_st_add( dst0, d44, 8 ); __insn_st_add( dst0, d44, 8 );
            __insn_st_add( dst0, d55, 8 ); __insn_st_add( dst0, d55, 8 );
            __insn_st_add( dst0, d66, 8 ); __insn_st_add( dst0, d66, 8 );
            __insn_st_add( dst0, d77, 8 ); __insn_st_add( dst0, d77, 8 );

            __insn_st_add( dst1, d00, 8 ); __insn_st_add( dst1, d00, 8 );
            __insn_st_add( dst1, d11, 8 ); __insn_st_add( dst1, d11, 8 );
            __insn_st_add( dst1, d22, 8 ); __insn_st_add( dst1, d22, 8 );
            __insn_st_add( dst1, d33, 8 ); __insn_st_add( dst1, d33, 8 );
            __insn_st_add( dst1, d44, 8 ); __insn_st_add( dst1, d44, 8 );
            __insn_st_add( dst1, d55, 8 ); __insn_st_add( dst1, d55, 8 );
            __insn_st_add( dst1, d66, 8 ); __insn_st_add( dst1, d66, 8 );
            __insn_st_add( dst1, d77, 8 ); __insn_st_add( dst1, d77, 8 );

            __insn_st_add( dst2, d00, 8 ); __insn_st_add( dst2, d00, 8 );
            __insn_st_add( dst2, d11, 8 ); __insn_st_add( dst2, d11, 8 );
            __insn_st_add( dst2, d22, 8 ); __insn_st_add( dst2, d22, 8 );
            __insn_st_add( dst2, d33, 8 ); __insn_st_add( dst2, d33, 8 );
            __insn_st_add( dst2, d44, 8 ); __insn_st_add( dst2, d44, 8 );
            __insn_st_add( dst2, d55, 8 ); __insn_st_add( dst2, d55, 8 );
            __insn_st_add( dst2, d66, 8 ); __insn_st_add( dst2, d66, 8 );
            __insn_st_add( dst2, d77, 8 ); __insn_st_add( dst2, d77, 8 );

            __insn_st_add( dst3, d00, 8 ); __insn_st_add( dst3, d00, 8 );
            __insn_st_add( dst3, d11, 8 ); __insn_st_add( dst3, d11, 8 );
            __insn_st_add( dst3, d22, 8 ); __insn_st_add( dst3, d22, 8 );
            __insn_st_add( dst3, d33, 8 ); __insn_st_add( dst3, d33, 8 );
            __insn_st_add( dst3, d44, 8 ); __insn_st_add( dst3, d44, 8 );
            __insn_st_add( dst3, d55, 8 ); __insn_st_add( dst3, d55, 8 );
            __insn_st_add( dst3, d66, 8 ); __insn_st_add( dst3, d66, 8 );
            __insn_st_add( dst3, d77, 8 ); __insn_st_add( dst3, d77, 8 );

        }

        dst0 += dstadd;
        dst1 += dstadd;
        dst2 += dstadd;
        dst3 += dstadd;

    }

}


static void scale5( uint8_t *src, uint32_t *pal, int srcstride, uint32_t *dst, int dststride )
{
    int x, y;
    uint32_t dstadd;
    uint32_t *dst0,*dst1,*dst2,*dst3,*dst4;

    dst0 = dst;
    dst1 = dst +   dststride;
    dst2 = dst + 2*dststride;
    dst3 = dst + 3*dststride;
    dst4 = dst + 4*dststride;

    dstadd = dststride*5-SCREENWIDTH*5;

    for( y = 0; y < SCREENHEIGHT; y++ ) {

        for( x = 0; x < SCREENWIDTH; x += 8 ) {
            uint64_t dl,dr;
            uint64_t d00,d01,d11,d22,d23,d33,d44,d45,d55,d66,d67,d77;
            uint64_t srcw;

            srcw = __insn_ld_add( src, 8 );

            dl = pal[__insn_bfextu( srcw,  0,  7 )]; d00 = __insn_v4int_l( dl, dl );
            dr = pal[__insn_bfextu( srcw,  8, 15 )]; d11 = __insn_v4int_l( dr, dr ); d01 = __insn_v4int_l( dr, dl );
            dl = pal[__insn_bfextu( srcw, 16, 23 )]; d22 = __insn_v4int_l( dl, dl );
            dr = pal[__insn_bfextu( srcw, 24, 31 )]; d33 = __insn_v4int_l( dr, dr ); d23 = __insn_v4int_l( dr, dl );
            dl = pal[__insn_bfextu( srcw, 32, 39 )]; d44 = __insn_v4int_l( dl, dl );
            dr = pal[__insn_bfextu( srcw, 40, 47 )]; d55 = __insn_v4int_l( dr, dr ); d45 = __insn_v4int_l( dr, dl );
            dl = pal[__insn_bfextu( srcw, 48, 55 )]; d66 = __insn_v4int_l( dl, dl );
            dr = pal[__insn_bfextu( srcw, 56, 63 )]; d77 = __insn_v4int_l( dr, dr ); d67 = __insn_v4int_l( dr, dl );

            __insn_st_add( dst0, d00, 8 ); __insn_st_add( dst0, d00, 8 ); __insn_st_add( dst0, d01, 8 );
            __insn_st_add( dst0, d11, 8 ); __insn_st_add( dst0, d11, 8 );
            __insn_st_add( dst0, d22, 8 ); __insn_st_add( dst0, d22, 8 ); __insn_st_add( dst0, d23, 8 );
            __insn_st_add( dst0, d33, 8 ); __insn_st_add( dst0, d33, 8 );
            __insn_st_add( dst0, d44, 8 ); __insn_st_add( dst0, d44, 8 ); __insn_st_add( dst0, d45, 8 );
            __insn_st_add( dst0, d55, 8 ); __insn_st_add( dst0, d55, 8 );
            __insn_st_add( dst0, d66, 8 ); __insn_st_add( dst0, d66, 8 ); __insn_st_add( dst0, d67, 8 );
            __insn_st_add( dst0, d77, 8 ); __insn_st_add( dst0, d77, 8 );

            __insn_st_add( dst1, d00, 8 ); __insn_st_add( dst1, d00, 8 ); __insn_st_add( dst1, d01, 8 );
            __insn_st_add( dst1, d11, 8 ); __insn_st_add( dst1, d11, 8 );
            __insn_st_add( dst1, d22, 8 ); __insn_st_add( dst1, d22, 8 ); __insn_st_add( dst1, d23, 8 );
            __insn_st_add( dst1, d33, 8 ); __insn_st_add( dst1, d33, 8 );
            __insn_st_add( dst1, d44, 8 ); __insn_st_add( dst1, d44, 8 ); __insn_st_add( dst1, d45, 8 );
            __insn_st_add( dst1, d55, 8 ); __insn_st_add( dst1, d55, 8 );
            __insn_st_add( dst1, d66, 8 ); __insn_st_add( dst1, d66, 8 ); __insn_st_add( dst1, d67, 8 );
            __insn_st_add( dst1, d77, 8 ); __insn_st_add( dst1, d77, 8 );

            __insn_st_add( dst2, d00, 8 ); __insn_st_add( dst2, d00, 8 ); __insn_st_add( dst2, d01, 8 );
            __insn_st_add( dst2, d11, 8 ); __insn_st_add( dst2, d11, 8 );
            __insn_st_add( dst2, d22, 8 ); __insn_st_add( dst2, d22, 8 ); __insn_st_add( dst2, d23, 8 );
            __insn_st_add( dst2, d33, 8 ); __insn_st_add( dst2, d33, 8 );
            __insn_st_add( dst2, d44, 8 ); __insn_st_add( dst2, d44, 8 ); __insn_st_add( dst2, d45, 8 );
            __insn_st_add( dst2, d55, 8 ); __insn_st_add( dst2, d55, 8 );
            __insn_st_add( dst2, d66, 8 ); __insn_st_add( dst2, d66, 8 ); __insn_st_add( dst2, d67, 8 );
            __insn_st_add( dst2, d77, 8 ); __insn_st_add( dst2, d77, 8 );

            __insn_st_add( dst3, d00, 8 ); __insn_st_add( dst3, d00, 8 ); __insn_st_add( dst3, d01, 8 );
            __insn_st_add( dst3, d11, 8 ); __insn_st_add( dst3, d11, 8 );
            __insn_st_add( dst3, d22, 8 ); __insn_st_add( dst3, d22, 8 ); __insn_st_add( dst3, d23, 8 );
            __insn_st_add( dst3, d33, 8 ); __insn_st_add( dst3, d33, 8 );
            __insn_st_add( dst3, d44, 8 ); __insn_st_add( dst3, d44, 8 ); __insn_st_add( dst3, d45, 8 );
            __insn_st_add( dst3, d55, 8 ); __insn_st_add( dst3, d55, 8 );
            __insn_st_add( dst3, d66, 8 ); __insn_st_add( dst3, d66, 8 ); __insn_st_add( dst3, d67, 8 );
            __insn_st_add( dst3, d77, 8 ); __insn_st_add( dst3, d77, 8 );

            __insn_st_add( dst4, d00, 8 ); __insn_st_add( dst4, d00, 8 ); __insn_st_add( dst4, d01, 8 );
            __insn_st_add( dst4, d11, 8 ); __insn_st_add( dst4, d11, 8 );
            __insn_st_add( dst4, d22, 8 ); __insn_st_add( dst4, d22, 8 ); __insn_st_add( dst4, d23, 8 );
            __insn_st_add( dst4, d33, 8 ); __insn_st_add( dst4, d33, 8 );
            __insn_st_add( dst4, d44, 8 ); __insn_st_add( dst4, d44, 8 ); __insn_st_add( dst4, d45, 8 );
            __insn_st_add( dst4, d55, 8 ); __insn_st_add( dst4, d55, 8 );
            __insn_st_add( dst4, d66, 8 ); __insn_st_add( dst4, d66, 8 ); __insn_st_add( dst4, d67, 8 );
            __insn_st_add( dst4, d77, 8 ); __insn_st_add( dst4, d77, 8 );

        }

        dst0 += dstadd;
        dst1 += dstadd;
        dst2 += dstadd;
        dst3 += dstadd;
        dst4 += dstadd;

    }

}

// This should never be used since we don't support 1920x1200.
// If 1920x1200 support is added, make a custom x6 scaler, please.
static void scale_fact( uint8_t *src, uint32_t *pal, int srcstride, uint32_t *dst, int dststride, int fact )
{
    int x, y, xx, yy;

    for( y = 0; y < SCREENHEIGHT; y++ ) {

        for( x = 0; x < SCREENWIDTH*fact; x += fact ) {

            uint32_t col = pal[*src++];
            uint32_t *dstrow = dst;

            for( yy = 0; yy < fact; yy++ ) {

                for( xx = 0; xx < fact; xx++ )  {

                    dstrow[x+xx] = col;

                }

                dstrow += dststride;

            }

        }

        dst += dststride*fact;

    }

}

// Quake is forced to run in 320x200 to avoid the water crash bug.
// So just scale here.. the downside is that gui is scaled too.
void VID_Update( vrect_t *rects )
{
    if( rects == NULL ) return;

    switch( multiply ) {
    case 1:
        scale1( vid_buffer, pal32, SCREENWIDTH, myscreen, mystride );
        break;

    case 2:
//        scale2_avg( vid_buffer, SCREENWIDTH, myscreen, mystride );
        scale2_epx( vid_buffer, pal32, SCREENWIDTH, myscreen, mystride );
        break;

    case 3:
        scale3_epx( vid_buffer, pal32, SCREENWIDTH, myscreen, mystride );
        break;

    case 4:
        scale4( vid_buffer, pal32, SCREENWIDTH, myscreen, mystride );
        break;

    case 5:
        scale5( vid_buffer, pal32, SCREENWIDTH, myscreen, mystride );
        break;

    default:
        scale_fact( vid_buffer, pal32, SCREENWIDTH, myscreen, mystride, multiply );
        break;
    }

}

/*
================
D_BeginDirectRect
================
*/
void D_BeginDirectRect( int x, int y, byte *pbitmap, int width, int height )
{
//    printf( "begindirect: %d,%d w=%d, h=%d\n", x, y, width, height );
}


/*
================
D_EndDirectRect
================
*/
void D_EndDirectRect( int x, int y, int width, int height )
{
//    printf( "enddirect: %d,%d w=%d, h=%d\n", x, y, width, height );
}
